/* 
 * File:     World.cpp
 * Author:   NMM
 * Date:     12.03.2006
 * Modified: 27.03.2006
 *
 * Purpose: Definitionen der Klasse World
 */
 
#include "World.h"
#include "Utilities.h"
#include "Definitions.h"

//Funktionen der Klasse Simulation

World::World(){
  int i;
  PaeSheep *this_sheep;
  PaeWulf *this_wulf, *current_wulf;
    
  /*
   * Construct the terrain
   */
  terr = new PaeTerrain();                       
  terr->ScaleTerrain(-5.0, 5.0);
  terr->ReBuildNormals();


  /* 
   * Add sheep to the sheep list
   */
   
  firstsheep = NULL;
  CreateSheep(STARTING_SHEEP);
  
  /* 
   * Add wulf to the wulf list
   */
   
  firstwulf = NULL;
  CreateWulf(STARTING_WULF);
  
  /*
   * One experimental Stone
   */
  stone = new Stone();
  stone->SetPosition(20.0,30.0,2.0);

  cout << "World created!" << endl;

}


void World::CreateSheep(){

  CreateSheep(1);

}

void World::CreateSheep(int n){

    PaeSheep *new_sheep;
    PaeSheep *last_sheep;
    float x;
    float y;

	/*
	 * Keine Schafe => Bye-bye
	 */

	if(n <= 0 ){
	  return;
	}


	/*
	 * We look for the last sheep in the list
	 */
	last_sheep = NULL;
    new_sheep = firstsheep;
    while(new_sheep != NULL){
      last_sheep = new_sheep;
	  new_sheep = new_sheep->nextsheep;
    }


	/*
	 * last_sheep is now NULL or pointing to the last sheep
	 */
	if(last_sheep == NULL){
	  x = PaeRand(0.0,99.0);
	  y = PaeRand(0.0,99.0);
	  firstsheep = new PaeSheep(x,y,this);
	  last_sheep = firstsheep;
	  n--;
	}


	/*
	 * last_sheep is pointing to the last sheep
	 */

	for(int i=0;i<n;i++){
	  x = PaeRand(0.0,99.0);
	  y = PaeRand(0.0,99.0);
	  last_sheep->nextsheep = new PaeSheep(x,y,this);
	  last_sheep = last_sheep->nextsheep;
	}
}



void World::CreateWulf(){

  CreateWulf(1);

}

void World::CreateWulf(int n){

    PaeWulf *new_wulf;
    PaeWulf *last_wulf;
    float x;
    float y;

	/*
	 * Keine Schafe => Bye-bye
	 */

	if(n <= 0 ){
	  return;
	}


	/*
	 * We look for the last wulf in the list
	 */
	last_wulf = NULL;
    new_wulf = firstwulf;
    while(new_wulf != NULL){
      last_wulf = new_wulf;
	  new_wulf = new_wulf->nextwulf;
    }


	/*
	 * last_wulf is now NULL or pointing to the last wulf
	 */
	if(last_wulf == NULL){
	  x = PaeRand(0.0,99.0);
	  y = PaeRand(0.0,99.0);
	  firstwulf = new PaeWulf(x,y,this);
	  last_wulf = firstwulf;
	  n--;
	}


	/*
	 * last_wulf is pointing to the last wulf
	 */

	for(int i=0;i<n;i++){
	  x = PaeRand(0.0,99.0);
	  y = PaeRand(0.0,99.0);
	  last_wulf->nextwulf = new PaeWulf(x,y,this);
	  last_wulf = last_wulf->nextwulf;
	}
}





void World::Draw(int visual){
  PaeSheep *this_sheep;
  PaeWulf *this_wulf;
  int i;
  i=0;
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
    stone->Draw();
  glPopMatrix();

  this_sheep = firstsheep;
  while(this_sheep != NULL){       //sage jedem Schaf, dass es sich zeichnen soll
    i++;
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
      this_sheep->Draw(visual);
    glPopMatrix();
    this_sheep = this_sheep->nextsheep;
	
  }
  this_wulf = firstwulf;
  while(this_wulf != NULL){
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
      this_wulf->Draw(visual);
    glPopMatrix();
    this_wulf = this_wulf->nextwulf;
  }
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
    terr->Draw();
  glPopMatrix();

  

}









void World::Prepare(){
     PaeSheep *this_sheep;
     PaeWulf *this_wulf;
     this_sheep = firstsheep;
     while(this_sheep != NULL){
       this_sheep->Think();
       this_sheep = this_sheep->nextsheep;
     }  
     this_wulf = firstwulf;
     while(this_wulf != NULL){
       this_wulf->Think();
       this_wulf = this_wulf->nextwulf;
     }    
        
}

void World::Action(){
  float cur_pos[3];
  int invert_flag;
  PaeSheep *this_sheep;
  PaeWulf *this_wulf;

  terr->GrowFood();

  this_sheep = firstsheep;
  while(this_sheep != NULL){
    this_sheep->Move();
	
	/*
	 * If sheep leaves terrain, turn it around!
	 */
	invert_flag = false;
	this_sheep->GetPosition(cur_pos);
	if (cur_pos[0] < 0.0){
	  cur_pos[0] = 0.1;
	  invert_flag = true;
	}
	if (cur_pos[0] >= 99.0){
	  cur_pos[0] = 98.9;
	  invert_flag = true;
	}
	if (cur_pos[1] < 0.0){
	  cur_pos[1] = 0.1;
	  invert_flag = true;
	}
	if (cur_pos[1] >= 99.0){
	  cur_pos[1] = 98.9;
	  invert_flag = true;
	}
	if(invert_flag){
	  this_sheep->InvertDirection();
	}

	/*
	 * Put sheep to the right height
	 */
	cur_pos[2] = terr->GetHeight(cur_pos[0],cur_pos[1]);
	this_sheep->SetPosition(cur_pos[0],cur_pos[1],cur_pos[2]);
	
	/*
	 * next, please!
	 */
    this_sheep = this_sheep->nextsheep;
  }         
  this_wulf = firstwulf;
  while(this_wulf != NULL){
       this_wulf->Act();
       this_wulf = this_wulf->nextwulf;
  } 
}

float World::GetHeight(float x, float y){
      return terr->GetHeight(x,y);
}


Stone * World::GetNearestStone(PaeSheep* asking_sheep){
  return stone;
}

PaeSheep * World::GetNearestSheep(PaeSheep* asking_sheep){
  PaeSheep *this_sheep;
  PaeSheep *nearest_sheep;
  float min_distance,distance;
  
  this_sheep = firstsheep;
  nearest_sheep = firstsheep;
  
  
  min_distance = Distance(this_sheep,asking_sheep);
  while(this_sheep->nextsheep != NULL){
    this_sheep = this_sheep->nextsheep; 
    if(this_sheep == asking_sheep){
      continue;
    }

    distance = Distance(this_sheep,asking_sheep);
    if(distance < min_distance){
      min_distance = distance;
      nearest_sheep = this_sheep;
    }
  }
  return(nearest_sheep);    
}



PaeSheep **World::GetFiveNearestSheep(PaeSheep* sheep){
   float  dist[5] = {10000.0,10000.0,10000.0,10000,10000.0};
   PaeSheep **near_sheep;
   PaeSheep *this_sheep;
   
   int i,j;
   float d;
   
   near_sheep = new PaeSheep*[5];
   near_sheep[0] = NULL; 
   near_sheep[1] = NULL; 
   near_sheep[2] = NULL; 
   near_sheep[3] = NULL; 
   near_sheep[4] = NULL;
   
   this_sheep = firstsheep;
   
   while(this_sheep != NULL){
     if(this_sheep == sheep){
       this_sheep = this_sheep->nextsheep;
       continue;
     }
     d = Distance(this_sheep,sheep);
     for(i=0;i<5;i++){
       if(d < dist[i]){
         /*
          * Verschiebe nachfolgende Schafe nach hinten
          */
         for(j=4;j>i;j--){
           near_sheep[j] = near_sheep[j-1];
           dist[j] = dist[j-1];
         }   
         near_sheep[i] = this_sheep;
         dist[i] = d;
         break;
         }
     }
     this_sheep = this_sheep->nextsheep;
   }
   
   return(near_sheep);
}


float World::GetGras(float x, float y){
   return(terr->GetFood(x,y));
}

void World::EatGras(float x, float y){
  terr->ReduceFood(x,y,FOOD_PER_SHEEP);
}


float* World::nextSheep(PaeWulf *w){
float *sheep_pos, *pos;
float dist = 100000.0, temp;
  sheep_pos = new float[3];  //die position eines schafes
  pos = new float[3];        //die position des naechsten schafes wird hier gespeichert
  
  PaeSheep *this_sheep;
  
  this_sheep = firstsheep;
  
  //wenn es keine schafe mehr gibt werden die woelfe in die irre gelockt
  pos[0] = PaeRand(0.0,99.0);
  pos[1] = PaeRand(0.0,99.0);
  pos[2] = 0.0;
  
  while(this_sheep != NULL){
    this_sheep->GetPosition(sheep_pos);
    temp = (sheep_pos[0] - w->pos[0])*(sheep_pos[0] - w->pos[0]) + (sheep_pos[1] - w->pos[1])*(sheep_pos[1] - w->pos[1]);
    if(temp < dist){
      dist = temp;
      pos[0] = sheep_pos[0];
      pos[1] = sheep_pos[1];
      pos[2] = sheep_pos[2];
    }
    this_sheep = this_sheep->nextsheep;
  }
  
  return pos;
}
   

//toetet ein schaf an ungefaher der angagabenen position
bool World::KillSheep(float *p){
PaeSheep *this_sheep, *last_sheep;
float *pos, dist;

  pos = new float[3];
    
  last_sheep = firstsheep;
  if(last_sheep != NULL){
    this_sheep = last_sheep->nextsheep;
  }
  //das erste schaf wird separat ueberprueft
  last_sheep->GetPosition(pos);
  dist = (p[0] - pos[0]) * (p[0] - pos[0]) + (p[1] - pos[1]) * (p[1] - pos[1]);
  if(dist < JAGDABSTAND){
    firstsheep = last_sheep->nextsheep;
    delete last_sheep;
    delete pos;
    return true;
  }
  
  while(this_sheep != NULL){
    this_sheep->GetPosition(pos);
    dist = (p[0] - pos[0]) * (p[0] - pos[0]) + (p[1] - pos[1]) * (p[1] - pos[1]);
    if(dist < JAGDABSTAND){
      last_sheep->nextsheep = this_sheep->nextsheep;
      delete this_sheep;
      delete pos;
      return true;
    }
    last_sheep = this_sheep;
    this_sheep = this_sheep->nextsheep;
  }
  
  delete pos;
  return false;
}
